/*
 * Support for Squeak plugin Serial and MIDI support
 * Ned Konz, July 14, 2000
 * 
 * $Revision$
 *
 * $Log$
 */

#include "sq.h"

#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

/*** Imported Variables ***/
/* extern int successFlag; */

/*** Module variables ***/

/* portNum	0=/dev/ttyS0, etc. */
/* include 1 byte for NUL at end */
#define PORT_NAME_SIZE 11
static const char serialPortBaseName[] = "/dev/ttyS0";

/* stopBits	0=1.5, 1=1, 2=2 */
/* I don't know how to get 1.5 stop bits. Oh well. So you get 2 instead */
#define MAX_STOP_BITS 2
/* c_cflag definitions */
static const unsigned int stopBitsDecode[MAX_STOP_BITS+1] = { CSTOPB, 0, CSTOPB };

/* dataBits	number of bits per character (5..8) */
/* note that since CS5 is 0, you will get 5 data bits if you ask for 0..4
 * as well */
#define MAX_DATA_BITS 8
/* c_cflag definitions */
static const unsigned int dataBitsDecode[MAX_DATA_BITS+1] = {
	0, 0, 0, 0, 0, CS5, CS6, CS7, CS8
};

/* parityType	0=no, 1=odd, 2=even  */
#define MAX_PARITY_TYPE 2
/* c_cflag definitions */
static const unsigned int parityTypeDecode[MAX_PARITY_TYPE+1] = {
	0,				/* none */
	PARENB|PARODD,	/* odd */
	PARENB			/* even */
};

/* inFlowCtrl	0=none, 1=XOn/XOff, 2=hardware handshaking
 * outFlowCtrl	0=none, 1=XOn/XOff, 2=hardware handshaking */
#define MAX_FLOW_CTRL 2

/* must be <= 10, because of 1-digit filename generation */
#define MAX_SERIAL_PORTS 10
static int serialFileDescriptors[MAX_SERIAL_PORTS];	/* index=squeak port# */
static struct termios savedSerialTermios[MAX_SERIAL_PORTS];	/* index=squeak port# */

/* dataRate	rate in bps */
typedef struct drDecode { int dataRate; speed_t code; } drDecode;
static drDecode dataRateDecode[] = {
	{ 0, B0 },		/* hang up */
	{ 50, B50 },
	{ 75, B75 },
	{ 110, B110 },
	{ 134, B134 },
	{ 150, B150 },
	{ 200, B200 },
	{ 300, B300 },
	{ 600, B600 },
	{ 1200, B1200 },
	{ 1800, B1800 },
	{ 2400, B2400 },
	{ 4800, B4800 },
	{ 9600, B9600 },
	{ 19200, B19200 },
	{ 38400, B38400 },
	{ 57600, B57600 },	/* below here may not be supported on your machine */
	{ 115200, B115200 },
	{ 230400, B230400 },
	{ 460800, B460800 },
	{ 500000, B500000 },	/* maybe OK for MIDI? */
	{ 576000, B576000 },
	{ 921600, B921600 },
	{ 1000000, B1000000 },	/* maybe OK for MIDI? */
	{ 1152000, B1152000 },
	{ 1500000, B1500000 },
	{ 2000000, B2000000 },	/* maybe OK for MIDI? */
	{ 2500000, B2500000 },
	{ 3000000, B3000000 },
	{ 3500000, B3500000 },
	{ 4000000, B4000000 },
	{ -1, B0 }			/* end marker */
};

/* This is the default setting for a termios structure on open */
static struct termios defaultTermios;

/*** Private Functions ***/

/* return the speed_t corresponding to the given data rate in bps,
 * or B0 if not found */
static speed_t serialDecodeSpeed(int speed)
{
	drDecode* p;
	for( p = dataRateDecode; p->dataRate >= 0; p++)
	{
		if (p->dataRate == speed)
			return p->code;
	}
	return B0;
}

/*** Public Functions ***/

/* return value ignored */
int serialPortClose(int portNum)
{
	int fd;
	if (portNum < 0 || portNum >= MAX_SERIAL_PORTS )
	{
		success(false);
		return 0;
	}
	/* Squeak wants to close already-closed ports... */
	if ((fd = serialFileDescriptors[ portNum ]) < 0)
	{
		success(true);
		return 0;
	}
	serialFileDescriptors[ portNum ] = -1;
	if (tcsetattr(fd, TCSAFLUSH, savedSerialTermios + portNum)
		|| close(fd))
	{
		success(false);
		return 0;
	}
	success(true);
	return 0;
}

/* Open the given serial port using the given settings. The data rate can be
 * any number that is in the table above; the driver is not as flexible
 * about the speed as the Mac driver, apparently. */
/* return value ignored */
int serialPortOpen(
		int portNum,
		int dataRate,
		int stopBitsType,
		int parityType,
		int dataBits,
		int inFlowCtrl,
		int outFlowCtrl,
		int xOnChar,
		int xOffChar)
{
	int fd;
	char serialPortName[ PORT_NAME_SIZE ];
	speed_t speed = serialDecodeSpeed( dataRate );
	struct termios flags;

	/* validate arguments */
	if (portNum < 0 || portNum >= MAX_SERIAL_PORTS
		|| speed == B0
		|| stopBitsType < 0 || stopBitsType > MAX_STOP_BITS
		|| parityType < 0 || parityType > MAX_PARITY_TYPE
		|| dataBits < 0 || dataBits > MAX_DATA_BITS
		|| inFlowCtrl < 0 || inFlowCtrl > MAX_FLOW_CTRL
		|| outFlowCtrl < 0 || outFlowCtrl > MAX_FLOW_CTRL
		|| (( inFlowCtrl == 1 || outFlowCtrl == 1 )
				&& ( xOnChar < 0 || xOnChar > 255
					|| xOffChar < 0 || xOffChar > 255 )))
	{
		success(false);
		return 0;
	}

	/* generate a filename (with last digit set to port number) */
	strcpy(serialPortName, serialPortBaseName);
	serialPortName[ PORT_NAME_SIZE - 2 ] = portNum + '0';

	/* open the device */
	if ((fd = open(serialPortName, O_RDWR|O_NONBLOCK|O_NOCTTY)) < 0)
	{
		success(false);
		return 0;
	}
	/* save the file descriptor */
	serialFileDescriptors[ portNum ] = fd;

	/* save the old state */
	if (tcgetattr(fd, savedSerialTermios + portNum))
	{
		success(false);
		return 0;
	}

	/* set up the new modes */
	flags = defaultTermios;

	/* input & output data rate */
	cfsetispeed(&flags, speed);
	cfsetospeed(&flags, speed);

	/* stop bits */
	flags.c_cflag &= ~CSTOPB;
	flags.c_cflag |= stopBitsDecode[ stopBitsType ];

	/* parity */
	flags.c_cflag &= ~(PARENB|PARODD);
	flags.c_cflag |= parityTypeDecode[ parityType ];

	/* data bits */
	flags.c_cflag &= ~CSIZE;
	flags.c_cflag |= dataBitsDecode[ dataBits ];

	/* flow control characters */
	if (inFlowCtrl == 1 || outFlowCtrl == 1)
	{
		flags.c_cc[ VSTART ] = xOnChar;
		flags.c_cc[ VSTOP ] = xOffChar;
	}

	flags.c_iflag &= ~(IXON|IXOFF|IXANY);
	flags.c_cflag &= ~CRTSCTS;

	if (inFlowCtrl == 1) flags.c_iflag |= IXOFF;
	if (outFlowCtrl == 1) flags.c_iflag |= IXON;
	if (inFlowCtrl == 2 || outFlowCtrl == 2) flags.c_cflag |= CRTSCTS;

	if (tcsetattr(fd, TCSANOW, &flags))	/* set it NOW */
	{
		success(false);
		return 0;
	}

	success(true);
	return 0;
}

/* Read up to count bytes from the given serial port into the given byte array.
   Read only up to the number of bytes in the port's input buffer; if fewer bytes
   than count have been received, do not wait for additional data to arrive.
   Return zero if no data is available, else number of bytes read */
int serialPortReadInto(int portNum, int count, int startPtr)
{
	int fd;
	ssize_t bytesRead;
	void* buffer = (void*)startPtr;		/* ints as pointers?? */

	if (portNum < 0
		|| portNum >= MAX_SERIAL_PORTS
		|| (fd = serialFileDescriptors[ portNum ]) < 0)
	{
		success(false);
		return 0;
	}

	bytesRead = read(fd, buffer, (size_t)count);
	if (bytesRead == (ssize_t)-1)
	{
		if (errno == EAGAIN)
			bytesRead = 0;
		else
		{
			success(false);
			return 0;
		}
	}

	success(true);
	return (int)bytesRead;
}

/* Write count bytes from the given byte array to the given serial port's
   output buffer. Return the number of bytes written. This implementation is
   synchronous: it doesn't return until the data has been sent. However, other
   implementations may return before transmission is complete. */
int serialPortWriteFrom(int portNum, int count, int startPtr)
{
	int fd;
	int bytesWritten;
	char* buffer = (void*)startPtr;

	if (portNum < 0
		|| portNum >= MAX_SERIAL_PORTS
		|| (fd = serialFileDescriptors[ portNum ]) < 0)
	{
		success(false);
		return 0;
	}

	bytesWritten = write(fd, buffer, (size_t)count);
	if (bytesWritten == (ssize_t)-1)
	{
		success(false);
		return 0;
	}

	success(true);
	return bytesWritten;
}

/* return true on success */
int serialPortInit(void)
{
	int i;

	/* initialize the file descriptors to invalid */
	for (i = 0; i < MAX_SERIAL_PORTS; i++)
		serialFileDescriptors[ i ] = -1;

	/* initialize our default termios structure (already 0'd) */
	defaultTermios.c_iflag = IGNBRK | IGNPAR;	/* ignore break, parity/framing errs */
	/* tcflag_t c_oflag output modes */
	/* defaultTermios.c_oflag = 0; */
	/* tcflag_t c_cflag control modes */
	/* defaultTermios.c_cflag = 0; */
	/* tcflag_t c_lflag local modes */
	/* defaultTermios.c_lflag = 0; */
	/* cc_t c_cc[NCCS] control chars */
	defaultTermios.c_cc[VTIME] = 5;
	defaultTermios.c_cc[VMIN] = 1;

	success(true);
	return 1;
}

/* return true on success */
int serialPortShutdown(void)
{
	success(true);
	return 1;
}

/* Put the given port into MIDI mode, which uses a clock supplied
   by an external MIDI interface adaptor to determine the data rate.
   Possible external clock rates: 31.25 KHz, 0.5 MHz, 1 MHz, or 2 MHz. */
int serialPortMidiClockRate(int portNum, int interfaceClockRate)
{
	success(false);
	return 0;
}

/* Close the given MIDI port. Do nothing if the port is not open.
   Fail if there is no port of the given number.*/
int sqMIDIClosePort(int portNum)
{
	success(false);
	return 0;
}

/* Return the current value of the clock used to schedule MIDI events.
   The MIDI clock is assumed to wrap at or before half the maximum
   positive SmallInteger value. This allows events to be scheduled
   into the future without overflowing into LargePositiveIntegers. 
   This implementation does not support event scheduling, so it
   just returns the value of the Squeak millisecond clock. */
int sqMIDIGetClock(void)
{
	success(false);
	return 0;
}

/* Return the number of available MIDI interfaces, including both
   hardware ports and software entities that act like ports. Ports
   are numbered from 0 to N-1, where N is the number returned by this
   primitive. */
int sqMIDIGetPortCount(void)
{
	success(false);
	return 0;
}

/* Return an integer indicating the directionality of the given
   port where: 1 = input, 2 = output, 3 = bidirectional. Fail if
   there is no port of the given number. */
int sqMIDIGetPortDirectionality(int portNum)
{
	success(false);
	return 0;
}

/* Copy the name of the given MIDI port into the string at the given
   address. Copy at most length characters, and return the number of
   characters copied. Fail if there is no port of the given number.*/
int sqMIDIGetPortName(int portNum, int namePtr, int length)
{
	success(false);
	return 0;
}

/* Open the given port, if possible. If non-zero, readSemaphoreIndex
   specifies the index in the external objects array of a semaphore
   to be signalled when incoming MIDI data is available. Note that
   not all implementations support read semaphores (this one does
   not); see sqMIDICanUseSemaphore. The interfaceClockRate parameter
   specifies the clock speed for an external MIDI interface
   adaptor on platforms that use such adaptors (e.g., Macintosh).
   Fail if there is no port of the given number.*/
int sqMIDIOpenPort(int portNum, int readSemaIndex, int interfaceClockRate)
{
	success(false);
	return 0;
}

/* Read or write the given MIDI driver parameter. If modify is 0,
   then newValue is ignored and the current value of the specified
   parameter is returned. If modify is non-zero, then the specified
   parameter is set to newValue. Note that many MIDI driver parameters
   are read-only; attempting to set one of these parameters fails.
   For boolean parameters, true = 1, false = 0. */
int sqMIDIParameter(int whichParameter, int modify, int newValue)
{
	success(false);
	return 0;
}

/* bufferPtr is the address of the first byte of a Smalltalk
   ByteArray of the given length. Copy up to (length - 4) bytes
   of incoming MIDI data into that buffer, preceded by a 4-byte
   timestamp in the units of the MIDI clock, most significant byte
   first. Implementations that do not support timestamping of
   incoming data as it arrives (see sqMIDIHasInputClock) simply
   set the timestamp to the value of the MIDI clock when this
   function is called. Return the total number of bytes read,
   including the timestamp bytes. Return zero if no data is
   available. Fail if the buffer is shorter than five bytes,
   since there must be enough room for the timestamp plus at
   least one data byte. */
int sqMIDIPortReadInto(int portNum, int count, int bufferPtr)
{
	success(false);
	return 0;
}

/* bufferPtr is the address of the first byte of a Smalltalk
   ByteArray of the given length. Send its contents to the given
   port when the MIDI clock reaches the given time. If time equals
   zero, then send the data immediately. Implementations that do
   not support a timestamped output queue, such as this one, always
   send the data immediately; see sqMIDIHasBuffer. */
int sqMIDIPortWriteFromAt(int portNum, int count, int bufferPtr, int time)
{
	success(false);
	return 0;
}


int midiInit(void)
{
	success(false);
	return 0;
}

int sqMIDIParameterGet(int which)
{
	success(false);
	return 0;
}

int sqMIDIParameterSet(int which, int value)
{
	success(false);
	return 0;
}

int midiShutdown(void)
{
	success(false);
	return 0;
}

/* vim: set ts=4 sw=4 cindent: */
